/* pvc_image.c
 * 
 * 
 * Copyright (C) 2012 Ivo Ihrke
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#include "../Image/pvc_image.h"


/* forward declarations for helper functions */
#ifdef HAVE_ZLIB

#include <zlib.h>

#else

#define gzFile     FILE*
#define gzgetc     getc
#define gzgets( file, pointer, numbytes )  fgets( pointer, numbytes, f )
#define gzopen     fopen
#define gzclose    fclose  

#endif

char   gz_skip_whitespace_and_comments( gzFile f );
char      skip_whitespace_and_comments( FILE *f );

float *load_pnmfile(const char *name, int *resx, int *resy);
char **load_pnm_supported_extensions(int *num_ext);
int    save_ppmfile( float *img, int xres, int yres, const char *name);




PVC_Image::PVC_Image()
{
  data = NULL;
  alignment_offset = 0;

  width = 0;
  height = 0;
}


PVC_Image::PVC_Image( int img_width, int img_height )
{
  /* memory for ARGB data */
  alloc_mem( width * height * 4 );

  width = img_width;
  height = img_height;
 
}

PVC_Image::~PVC_Image()
{
  if ( data != NULL )
    {
      /* get the original data pointer */
      data = (float *) ( (unsigned char *)data - alignment_offset );

      /* free it */
      delete [] data;
    }
}


int PVC_Image::resize( int img_width, int img_height, int mode )
{
  /* remove old image */
  if ( data != NULL )
    {
      /* get the original dat pointer */
      data -= alignment_offset;

      /* free it */
      delete [] data;
    }

  /* memory for ARGB data */
  alloc_mem( img_width * img_height * 4, mode );

  width = img_width;
  height = img_height;
  
  return 0;
}


/* !
   loads an image:
   - values are in [0,1], 
   - ARGB interlaced order = 4 floats per pixel
   - the upper left pixel is the first data element
   - storage is row-major

   returns negative number on failure 
*/
int PVC_Image::load( const char *filename )
{
  float *ret = NULL;

  /* cannot load noname image */
  if ( filename == NULL )
    return -1;
  
  /* load_pnmfile allocates memory for PVC_Image.data */
  ret = load_pnmfile( filename, &width, &height );

 /* error loading image, indicate by returning NULL */
  if ( ret == NULL )
    {
      free ( ret );
      return -1;
    }

  /* get memory */
  resize( width, height );


  /* copy data */
  for ( int i = 0; i < width * height * 4; i++ )
    data [ i ] = ret [ i ];

  /* load_pnmfile is an ANSI C code */
  free ( ret );
 
  return 0;
}

/* saves an image, returns -1 on error */
int PVC_Image::save( const char *filename )
{
  int err = 0;

  if ( data == NULL || 
       filename == NULL )
    return -1;
  
  err = save_ppmfile( data, width, height, filename );

  return err;
}

/* allocates memory in data */
int PVC_Image::alloc_mem( unsigned long int num_bytes, int mode )
{
  if( data != NULL )
    delete [] data;

  data = new float [ num_bytes + 15 ];

  if ( mode == PVC_IMAGE_ALIGNED_MEM )
    // the last % 16 is for the case that the memory is already aligned 
    alignment_offset = (16 - (unsigned long int)data % 16) % 16; 

  if ( mode == PVC_IMAGE_UNALIGNED_MEM )
    alignment_offset = 0;
 
  data = (float *)((unsigned char*)data + alignment_offset);

#ifdef DEBUG
  fprintf( stderr,"PVC_Image::alloc_mem: Memory is aligned at %d\n", (unsigned long int)data % 16 );
#endif
  
  return 0;
}




/* -------------------------------------------------------------------- */
/* ------------------------ helper functions -------------------------- */
/* -------------------------------------------------------------------- */

/*! gzipped files - returns first non-whitespace or comment character */
char gz_skip_whitespace_and_comments( gzFile f )
{
  char c;

  /* skip whitespace */
 again:

  while ( isspace ( c = gzgetc ( f ) ) );

  if ( c == '#' )
    {
      /*skip comment*/
      while( ( c = gzgetc ( f ) ) != '\n' );
    }
  
  if ( isspace ( c ) )
    goto again;
  
  return c;
}




#define BIT16_ADD            100

#define TYPE_RGB_RAW_8       6             /* P6 maxval<=255 */
#define TYPE_RGB_RAW_16      6+BIT16_ADD   /* P6 maxval >255 */
#define TYPE_RGB_ASCII       3             /* P3 */

#define TYPE_GRAY_RAW_8      5             /* P5 maxval<=255 */
#define TYPE_GRAY_RAW_16     5+BIT16_ADD   /* P5 maxval >255 */
#define TYPE_GRAY_ASCII      2             /* P2 */

#define TYPE_BM_RAW          4             /* P4 */
#define TYPE_BM_ASCII        1             /* P1 */  


#define PNM_NUM_EXTENSIONS   6

char pnm_extensions [ PNM_NUM_EXTENSIONS ][ 8 ] = {".pbm",".pgm",".ppm",".pbm.gz",".pgm.gz",".ppm.gz"};




char **load_pnm_supported_extensions(int *num_ext)
{
  char **ret;
  int i;

  ret=(char **)malloc(PNM_NUM_EXTENSIONS*sizeof(char *));
  *num_ext=PNM_NUM_EXTENSIONS;

  for (i=0;i<PNM_NUM_EXTENSIONS;i++)
    ret[i]=pnm_extensions[i];

  return ret;
}





float *load_pnmfile(const char *name, int *outx, int *outy)
{
  gzFile f;
  int    type;
  int    max_val;
  char   c;

  float *ret_data = NULL;
  int    r, g, b;
  int    i;
  int    lbit          = 0;
  int    shift         = 0;
  int    bm_shift      = 0;
  int    pixel_counter = 0;

  char buffer  [ 2000 ];
  char buffer2 [ 2000 ];
  
#ifdef DEBUG
  fprintf(stderr,"load_pnmfile: opening '%s'\n",name);
#endif

#if 0
  if ( ( strncmp( name, "-.p", 3 ) == 0 ) && 
       ( name [ 3 ] == 'b' || name [ 3 ] == 'g' || name [ 3 ] == 'p' ) &&
       ( strlen( name ) == 5 ) )
    {

      /* setmode( fileno( stdin ), O_BINARY ); */

      f = gzdopen( dup ( fileno( stdin ) ), "rb" );

#ifdef DEBUG
      fprintf( stderr,"load_pnmfile: opening stdin\n" );
#endif

    }
  else
#endif

    f = gzopen(name,"rb");  // rb is need for windows

  if( f == NULL )
    {
      fprintf(stderr,"load_pnmfile: Could not open %s\n",name);

      return NULL;
      /*       exit(1); */
    }

#ifdef DEBUG
  fprintf(stderr,"load_pnmfile: '%s' opened \n",name);
#endif  

  buffer [ 0 ] = gz_skip_whitespace_and_comments ( f );
  gzgets ( f, buffer + 1, 1000 );

  sscanf( buffer, "P%d", &type);

  buffer [ 0 ] = gz_skip_whitespace_and_comments ( f );

  gzgets ( f, buffer + 1, 1000 );

  sscanf( buffer,"%d",outx);

  sprintf ( buffer2, "%d", *outx );
  i = strlen ( buffer2 );

/*   buffer [ 0 ] = gz_skip_whitespace_and_comments ( f ); */
/*   gzgets ( f, buffer + 1, 1000 ); */

  sscanf( buffer + i + 1, "%d", outy );

#ifdef DEBUG
  fprintf(stderr,"load_pnmfile: read header \n" );
#endif

  if ( ( type != TYPE_BM_RAW ) && 
       ( type != TYPE_BM_ASCII ) )
    {
      buffer [ 0 ] = gz_skip_whitespace_and_comments ( f );
      gzgets ( f, buffer + 1, 1000 );

      sscanf ( buffer, "%d", &max_val);

      max_val > 255 ? type += BIT16_ADD : type;

      /* find most significant bit in max_val (to fit values in 0..255 range */
      for ( i = 0; i < 16; i++ )
	 ( (max_val >> i ) & 1 ) ? lbit = i : lbit;
      shift = ( 0 > ( lbit + 1 ) - 8 ) ? 0 : ( lbit + 1 ) - 8; 
  
    }
  else
    {
      shift = 0;
      bm_shift = 0;
      max_val = 1;
    }

#ifdef DEBUG
  fprintf(stderr,"load_pnmfile:  Type: %d\nload_pnmfile: Res: %dx%d (Max: %d - Shift: %d)\n",type,*outx,*outy,max_val,shift);
#endif

  /* skip single whitespace - gzgets scanned this already - filepointer is now at start of image data */
/*   if ( type != TYPE_BM_RAW ) */
/*     buffer [ 0 ] = gzgetc( f );    */


  /* alloc image mem */
  ret_data = (float *) malloc (*outx * *outy * sizeof(float) * 4);

#ifdef DEBUG
  fprintf(stderr,"load_pnmfile: allocated memory ...\n" );
#endif

  /* read image data */
  if ( ret_data != NULL )
    for ( i = 0; i < *outx * *outy; i++ )
      {
	switch(type)
	  {
	case TYPE_RGB_RAW_8  :r = (unsigned char) gzgetc( f );
	                      g = (unsigned char) gzgetc( f );
			      b = (unsigned char) gzgetc( f );
			      break;
	case TYPE_RGB_RAW_16 :r = ((unsigned char) gzgetc( f ) << 8 )+
			           (unsigned char) gzgetc( f );
	                      g = ((unsigned char) gzgetc( f ) << 8 )+
			           (unsigned char) gzgetc( f );
			      b = ((unsigned char) gzgetc( f ) << 8 )+
			           (unsigned char) gzgetc( f );
			      break;
	case TYPE_RGB_ASCII  :gzgets ( f, buffer + 1, 1000 );
 	                      sscanf ( buffer, "%d", &r );
                 	      buffer [ 0 ] = gz_skip_whitespace_and_comments ( f );
			      gzgets ( f, buffer + 1, 1000 );
 	                      sscanf ( buffer, "%d", &g );
                 	      buffer [ 0 ] = gz_skip_whitespace_and_comments ( f );
			      gzgets ( f, buffer + 1, 1000 );
 	                      sscanf ( buffer, "%d", &b );
                 	      buffer [ 0 ] = gz_skip_whitespace_and_comments ( f );
			      break;
	case TYPE_GRAY_RAW_8 :r = (unsigned char) gzgetc ( f );
	                      g = r; 
			      b = r;
                              break;
	case TYPE_GRAY_RAW_16:r = ((unsigned char) gzgetc ( f ) << 8 )+
				  ((unsigned char) gzgetc ( f ) );
	                      g = r;
			      b = r;
                              break;
	case TYPE_GRAY_ASCII :
 	                      buffer [ 0 ] = gz_skip_whitespace_and_comments ( f );
			      gzgets ( f, buffer + 1, 1000 );
 	                      sscanf ( buffer, "%d", &r );
			      /*
			      fscanf(f,"%d",&r);
	                      skip_whitespace_and_comments(f);
			      */
	                      g = r;
			      b = r;
                              break;
	case TYPE_BM_RAW     : if ( pixel_counter == *outx )
	                         {
				   pixel_counter = 0;
				   bm_shift = 0;
				 }
	                      (bm_shift==0)?c=gzgetc(f):c;
	                      r=(c&(1<<(7-bm_shift)))?0:1;
                              g=r;b=r;
                              bm_shift=(bm_shift+1)&7;
			      pixel_counter++;
			      break;
	case TYPE_BM_ASCII   :
                 	      buffer [ 0 ] = gz_skip_whitespace_and_comments ( f );
			      gzgets ( f, buffer + 1, 1000 );
 	                      sscanf ( buffer, "%d", &r );
			      
			      /*
	                      fscanf(f,"%d",&r);
	                      skip_whitespace_and_comments(f);
			      */
			      
	                      g = ( r == 0 ) ? 1 : 0;
			      b = g;
			      r = g;
                              break;
	}
      
      /*  r=r>>shift; */
      /*  g=g>>shift; */
      /*  b=b>>shift; */
      
      /*  image[i]=(r<<16)+(g<<8)+b; */
      /* little endian ARGB */
      ret_data [ i * 4 + 3 ] = 1.0f;  /* completely opaque */
      ret_data [ i * 4 + 2 ] = (float)r / (float)max_val;  
      ret_data [ i * 4 + 1 ] = (float)g / (float)max_val;  
      ret_data [ i * 4 + 0 ] = (float)b / (float)max_val;  

      if (r > max_val || g > max_val || b > max_val || r <0 || g < 0 || b < 0)
	{
	  fprintf(stderr,"load_pnmfile: PNM - file not consistent, RGB (%d,%d,%d) at least one component greater than max_val (%d)\n",r,g,b,max_val);
	}
  
      }
  else
    {
      perror("load_pnmfile: memory allocation failed");
    }

#ifdef DEBUG
  fprintf( stderr,"load_pnmfile: closing file desc.\n" );
#endif

  gzclose(f);
  
#ifdef DEBUG
  fprintf( stderr,"load_pnmfile: returning\n" );
#endif

  return ret_data;
}





#define PPM_NUM_EXTENSIONS   2

char ppm_extensions[PPM_NUM_EXTENSIONS][5] = {".pgm",".ppm"};




char **save_ppm_supported_extensions(int *num_ext)
{
  char **ret;
  int i;
  
  ret=(char **)malloc(PPM_NUM_EXTENSIONS*sizeof(char *));
  *num_ext=PPM_NUM_EXTENSIONS;

  for (i=0;i<PPM_NUM_EXTENSIONS;i++)
    ret[i]=ppm_extensions[i];

  return ret;
}






int save_ppmfile( float *img, int xres, int yres, const char *name)
{
  FILE *f;
  int i;
  int max_val = 65535;

  unsigned short int current;                                                                                                                                       
  float maxr = float(-10e10), minr = float(10e10);                                                                                                                                
  float maxg = float(-10e10), ming = float(10e10);                                                                                                                                
  float maxb = float(-10e10), minb = float(10e10);                                                                                                                                
  

#ifdef DEBUG
  fprintf ( stderr,"save_ppmfile: start ...\n" );
#endif

  if ( ( strncmp( name, "-.ppm", 5 ) == 0 ) && 
       ( strlen( name ) == 5 ) )
    {

#ifndef O_BINARY
      #define O_BINARY 0
#endif

      /* setmode( fileno( stdout ), O_BINARY );  */

      f = stdout; /* fopen( fileno( stdout ), "wb" ); */

      fprintf( stderr,"save_ppmfile:  opening stdout\n" );
    }
  else
    f = fopen( name, "wb" );
  
  if( f == NULL )
    {
      fprintf( stderr, "save_ppmfile:  Could not open %s for writing\n", name);
      exit ( 1 );
    }

                                                                                                                                                                
  fprintf(f,"P6\n#Lg_SavePPM\n%d %d\n%d\n", xres, yres, max_val );                                                                                         

  /* convert from float to 16bit */                                                                             
                                                                             
  for ( i = 0; i < xres * yres; i++ )
    {                                                                        
      /*   fwrite(img,1,3,f); */                                             
      current = (unsigned short int)(img[2] * (float)max_val);               
      fwrite((char *)(&current) + 1, ( size_t) 1, ( size_t) 1, f );          
      fwrite((char *)(&current)    , ( size_t) 1, ( size_t) 1,f);            
                                                                             
      current = (unsigned short int)(img[1] * (float)max_val);               
      fwrite((char *)(&current) + 1, ( size_t) 1, ( size_t) 1,f );           
      fwrite((char *)(&current)    , ( size_t) 1, ( size_t) 1,f );           
                                                                             
      current = (unsigned short int)(img[0] * (float)max_val);               
      fwrite((char *)(&current) + 1, ( size_t) 1, ( size_t) 1,f);            
      fwrite((char *)(&current)    , ( size_t) 1, ( size_t) 1,f);            
                                                                             
                                                                             
      maxr  = ( maxr > img[2] ) ? maxr : img[2];                             
      minr  = ( minr < img[2] ) ? minr : img[2];                             
                                                                             
      maxg  = ( maxg > img[1] ) ? maxg : img[1];                             
      ming  = ( ming < img[1] ) ? ming : img[1];                             
                                                                             
      maxb  = ( maxb > img[0] ) ? maxb : img[0];                             
      minb  = ( minb < img[0] ) ? minb : img[0];                             
                                                                             
                                                                             
      img+=4;                                                                
    }                                                                        
                                                                             
  if ( f != stdout )
    fclose( f );                                                                             
               
#ifdef DEBUG                                                              
  fprintf ( stderr,"save_ppmfile: R - Min: %0.2f Max: %0.2f\n", minr, maxr );              
  fprintf ( stderr,"save_ppmfile: G - Min: %0.2f Max: %0.2f\n", ming, maxg );              
  fprintf ( stderr,"save_ppmfile: B - Min: %0.2f Max: %0.2f\n", minb, maxb );              
#endif                                                                             
                                                                             
  
#ifdef DEBUG
  fprintf ( stderr,"save_ppmfile: done\n" );
#endif

  return 0;
}
