/* 

   small demo program, reading and writing an image


   (c) 2012  Ivo Ihrke
       Universitaet des Saarlandes
       ihrke@mmci.uni-saarland.de
 
*/

#include <stdlib.h>
#include <stdio.h>

#ifdef WIN32
#define _USE_MATH_DEFINES
#endif
#include <math.h>

#include <time.h>
#include <string.h>

#include <vector>
#include <iostream>
#include <cassert>
#include <limits>

#include <omp.h>

#ifndef WIN32
#include <unistd.h>
#endif

#include "../../Image/pvc_image.h"
#include "../../Image/pvc_image_gray.h"
#include "../../Image/filterimage.h"

using namespace std;

class OMP_Filter_Image : public Filter_Image {
  virtual Filter_Image *filterGauss( float SIGMA );
};

/* 
   use a separable kernel, remove branching in inner-most loop by offset lookup 

   this is slightly slower than the ugly version filter_conv_sep2D_nobranch

 */
Filter_Image *OMP_Filter_Image::filterGauss( float SIGMA )
{
  /* filter image */
  PVC_Kernel* kernel = new PVC_Kernel;
  int support= int(SIGMA * 3 * 2 +1);
  kernel -> resize( support, support );

  for ( float j = float(-kernel -> halfsize()); j <= float(kernel -> halfsize()); j += 1.0 )
    for ( float i = float(-kernel -> halfsize()); i <= float(kernel -> halfsize()); i += 1.0 )
      {
	//	fprintf (stderr, "Gauss (%d,%d) = %0.4f \n", (int)i, (int)j, 1.0 / ( 2.0 * SIGMA * SIGMA ) * exp( -( i * i  +  j * j ) / ( 2.0 * SIGMA * SIGMA ) ) );
	//kernel -> put_value( i, j, 1.0 / ( 2.0 * M_PI * SIGMA * SIGMA ) * exp( -( i * i  +  j * j ) / ( 2.0 * SIGMA * SIGMA ) ) );
	kernel -> put_value( (int)i, (int)j, 1.0f / ( sqrtf ( 2.0f * float(M_PI)) * float(SIGMA)) * expf( -( i * i ) / ( 2.0f * float(SIGMA) * float(SIGMA) ) ) );
      }

  Filter_Image *dum       = NULL;
  Filter_Image *ret       = NULL;

  int          *xoffset   = NULL;
  int          *yoffset   = NULL;

  float        *dumdata   = NULL;
  float        *retdata   = NULL;
  float        *  kdata   = NULL;

  int           hs        = kernel -> halfsize();

  /* create intermediate result image */
  dum = new OMP_Filter_Image();
  dum -> resize( w(), h() );
  dumdata = dum -> get_data();
  
  /* create output image */
  ret = new OMP_Filter_Image();
  ret -> resize( w(), h() );
  retdata = ret -> get_data();

  kdata = kernel -> get_data();

  // create lookup tables

  xoffset = new int [ w() + 2 * hs + 1 ];
  yoffset = new int [ h() + 2 * hs + 1 ];

  // left boundary 
  for ( int i = 0; i < hs; i++ )
    xoffset [ i ] = 0;

  // inner region
  for ( int i = hs; i < w() + hs; i++ )
    xoffset [ i ] = i - hs;

  // right boundary 
  for ( int i = w() + hs; i < w() + 2 * hs; i++ )
    xoffset [ i ] = w() - 1;


  // upper boundary 
  for ( int i = 0; i < hs; i++ )
    yoffset [ i ] = 0;

  // inner region
  for ( int i = hs; i < h() + hs; i++ )
    yoffset [ i ] = i - hs;

  // lower boundary 
  for ( int i = h() + hs; i < h() + 2 * hs; i++ )
    yoffset [ i ] = h() - 1;


  //---------------- HORIZONTAL 1D CONV -------------------
 
  // horizontal conv.
  for ( int j = 0; j < h(); j++ )     // image y-axis
    for ( int i = 0; i < w(); i++ )   // image x-axis
      {
	
	float sum [ 4 ];
	
	memset( sum, 0, 16 );

	//horizontal conv
	for ( int k = -hs; k <= hs; k++ )   // kernel x-axis
	  {
	    for ( int c = 0; c < 4; c++ )
	      {
		float  kernelval = kdata[ k + hs ]; //kernel -> get_value( k, 0 );
		float  imgval = data [ ( j * width  + xoffset [ i + k + hs ]  ) * 4 + c ]; 

		sum [ c ] += kernelval * imgval;
	      }
	  }

	memcpy( dumdata + ( j * width + i ) * 4, sum, 16 );
      }


  //---------------- VERTICAL 1D CONV -------------------


  // vertical conv
  for ( int j = 0; j < h(); j++ )     // image y-axis
    for ( int i = 0; i < w(); i++ )   // image x-axis
      {
	
	float sum [ 4 ];
	

	memset( sum, 0, 16 );

	//vertical conv
	for ( int l = -hs; l <= hs; l++ )   // kernel x-axis
	  {

	    for ( int c = 0; c < 4; c++ )
	      {
		// HACK use only center x-axis of kernel as 1D
		// kernel to avoid re-implementation of 1D kernel

		float  kernelval = kdata [ l + hs ];//kernel -> get_value( l, 0 );
		float  imgval    = dumdata [ ( yoffset [ j + l + hs ] * width  + i  ) * 4 + c ];

		sum [ c ] += kernelval * imgval;
	      }
	  }
	memcpy( retdata + ( j * w() + i ) * 4, sum, 16 );
      }


  delete [] xoffset;
  delete [] yoffset;

  
  delete dum;

  return ret;
}


class ScaleSpace
{
	private:
	int nOctaves;
	int nScalesPerOctave;
	float sigma0;
	int width0;
	int height0;
	vector< vector< Filter_Image* > > data;
	
	float calcSigma(int octId, int scaleId)
	{
		float sigmaNominal= 0.5;
		float sigma= sigma0 * pow(2, octId + (float) scaleId/nScalesPerOctave);
		return sqrt(sigma*sigma - sigmaNominal*sigmaNominal);
	}

	public:	
	ScaleSpace(int nOctaves= 6, int nScalesPerOctave= 5, float sigma0= 1.6) : nOctaves(nOctaves), nScalesPerOctave(nScalesPerOctave), sigma0(sigma0)
	{
		printf("constructed: %d %d %f\n", nOctaves, nScalesPerOctave, sigma0);
	}

	Filter_Image* getScale(int octId, int scaleId) {return data[octId][scaleId];};

	int getOctaves() { return nOctaves; }
	int getScalesPerOctave() { return nScalesPerOctave; }
	int getWidth0() { return width0; }
	int getHeight0() { return height0; }
	
	void process(Filter_Image* image)
	{
		width0= image->w();
		height0= image->h();
		data.resize(nOctaves);
		for (int octId= 0; octId<nOctaves; octId++)
		{
			cerr << "octave : " << octId << endl;
			data[octId].resize(nScalesPerOctave);
			int width= int(float(width0) / powf(2, float(octId)));
			int height= int(float(height0) / powf(2, float(octId)));
			for (int scaleId= 0; scaleId<nScalesPerOctave; scaleId++)
			{
				float sigma= calcSigma(octId, scaleId);	
				cerr << "sigma: " << sigma << " " << endl;
				data[octId][scaleId]= image->filterGauss(sigma);

	
			}
		}
	}

	// save all the scales space images, argument template is supposed to contain a formated string
	// like "image_%d_%d.ppm"
	void saveImages(char* filenameTemplate)
	{
		char filename[1024];
		for (int octId= 0; octId<nOctaves; octId++)
		{
			for (int scaleId= 0; scaleId<nScalesPerOctave; scaleId++)
			{
				sprintf(filename, filenameTemplate, octId, scaleId);
				data[octId][scaleId]->save(filename);
			}
		}
	}

};


class ScaleSpaceMax
{
	public:
	int x;
	int y;
	int octId;
	int scaleId;

	ScaleSpaceMax(int x, int y, int octId, int scaleId): x(x), y(y), octId(octId), scaleId(scaleId)
	{
	}
};

class DifferenceOfGaussians
{
	float threshold;
	float delta;
	float edgeThresh;
	vector< vector< PVC_Image_Gray* > > data;
	public:
	DifferenceOfGaussians(float threshold= 0.0, float delta= 0.0, float edgeThresh= 0.0) : threshold(threshold), delta(delta), edgeThresh(edgeThresh)
	{
	}

	void process(ScaleSpace& scaleSpace)
	{
		data.resize(scaleSpace.getOctaves());
		for( int octId= 0; octId<scaleSpace.getOctaves(); octId++)
		{
			data[octId].resize(scaleSpace.getScalesPerOctave()-1);
			for (int scaleId= 0; scaleId<scaleSpace.getScalesPerOctave()-1; scaleId++)
			{
				Filter_Image* im1= scaleSpace.getScale(octId, scaleId);
				Filter_Image* im2= scaleSpace.getScale(octId, scaleId+1);
				data[octId][scaleId]= im1->substractToGray(im2);
					
			}
		}		
	}
	
	// save all the scales space images, argument template is supposed to contain a formated string
	// like "image_%d_%d.ppm"
	// data is normalized to [0,1] ; changes the data!
	void saveImages(char* filenameTemplate)
	{
		char filename[1024];
		for (int octId= 0; octId<int(data.size()); octId++)
		{
			for (int scaleId= 0; scaleId<int(data[0].size()); scaleId++)
			{
				sprintf(filename, filenameTemplate, octId, scaleId);
				data[octId][scaleId]->normalize();	
				data[octId][scaleId]->save(filename);
			}
		}
	}

	inline float get_value(int octId, int scaleId, int x, int y)
	{
		return data[octId][scaleId]->get_value(x, y);
	}
							
	float computeEdgeScore(int octId, int scaleId, int x, int y)
	{
		// compute hessian
		float dx1= get_value(octId, scaleId, x+1, y) - get_value(octId, scaleId, x, y);
		float dx2= get_value(octId, scaleId, x, y) - get_value(octId, scaleId, x-1, y);
		float dxdx= dx1-dx2;
		float dy1= get_value(octId, scaleId, x, y+1) - get_value(octId, scaleId, x, y);
		float dy2= get_value(octId, scaleId, x, y) - get_value(octId, scaleId, x, y-1);
		float dydy= dy1-dy2;
		float dxdy= dx1-dy1;

		// trace of hessian
		float trace= dxdx + dydy;
		float determinant= dxdx * dydy - dxdy*dxdy;

		return fabs(trace*trace/determinant);

	}

	vector<ScaleSpaceMax> findMax()
	{
		vector< ScaleSpaceMax > scaleSpaceMax;
		int nOctaves= data.size();
		for (int octId= 0; octId<nOctaves; octId++)
		{
			int nScales= data[octId].size();
			for(int scaleId= 1; scaleId<nScales-1; scaleId++)
			{
				int width= data[octId][scaleId]->w();
				int height= data[octId][scaleId]->h();
				for (int y= 1; y<height-1; y++)
				{
					for (int x= 1; x<width; x++)
					{
						const float centerValueBase= get_value(octId, scaleId, x, y);
						if (fabs(centerValueBase) > threshold)
						{
						float centerValue= centerValueBase-delta;
						if (
							// previous scale
							get_value(octId, scaleId-1, x-1, y-1)<centerValue &&
							get_value(octId, scaleId-1, x, y-1)<centerValue &&
							get_value(octId, scaleId-1, x+1, y-1)<centerValue &&
							get_value(octId, scaleId-1, x-1, y)<centerValue &&
							get_value(octId, scaleId-1, x, y)<centerValue &&
							get_value(octId, scaleId-1, x+1, y)<centerValue &&
							get_value(octId, scaleId-1, x-1, y+1)<centerValue &&
							get_value(octId, scaleId-1, x, y+1)<centerValue &&
							get_value(octId, scaleId-1, x+1, y+1)<centerValue &&
							
							// this scale	
							get_value(octId, scaleId, x-1, y-1)<centerValue &&
							get_value(octId, scaleId, x, y-1)<centerValue &&
							get_value(octId, scaleId, x+1, y-1)<centerValue &&
							get_value(octId, scaleId, x-1, y)<centerValue &&
							get_value(octId, scaleId, x+1, y)<centerValue &&
							get_value(octId, scaleId, x-1, y+1)<centerValue &&
							get_value(octId, scaleId, x, y+1)<centerValue &&
							get_value(octId, scaleId, x+1, y+1)<centerValue &&
							
							// next scale
							get_value(octId, scaleId+1, x-1, y-1)<centerValue &&
							get_value(octId, scaleId+1, x, y-1)<centerValue &&
							get_value(octId, scaleId+1, x+1, y-1)<centerValue &&
							get_value(octId, scaleId+1, x-1, y)<centerValue &&
							get_value(octId, scaleId+1, x, y)<centerValue &&
							get_value(octId, scaleId+1, x+1, y)<centerValue &&
							get_value(octId, scaleId+1, x-1, y+1)<centerValue &&
							get_value(octId, scaleId+1, x, y+1)<centerValue &&
							get_value(octId, scaleId+1, x+1, y+1)<centerValue)
						{
							float edgeScore= computeEdgeScore(octId, scaleId, x, y);
							if (edgeScore < (edgeThresh+1)*(edgeThresh+1)/edgeThresh)
							{
								scaleSpaceMax.push_back(ScaleSpaceMax(x, y, octId, scaleId));	
							}
						}
						centerValue= centerValueBase+delta;
						if (
							// previous scale
							get_value(octId, scaleId-1, x-1, y-1)>centerValue &&
							get_value(octId, scaleId-1, x, y-1)>centerValue &&
							get_value(octId, scaleId-1, x+1, y-1)>centerValue &&
							get_value(octId, scaleId-1, x-1, y)>centerValue &&
							get_value(octId, scaleId-1, x, y)>centerValue &&
							get_value(octId, scaleId-1, x+1, y)>centerValue &&
							get_value(octId, scaleId-1, x-1, y+1)>centerValue &&
							get_value(octId, scaleId-1, x, y+1)>centerValue &&
							get_value(octId, scaleId-1, x+1, y+1)>centerValue &&
							
							// this scale	
							get_value(octId, scaleId, x-1, y-1)>centerValue &&
							get_value(octId, scaleId, x, y-1)>centerValue &&
							get_value(octId, scaleId, x+1, y-1)>centerValue &&
							get_value(octId, scaleId, x-1, y)>centerValue &&
							get_value(octId, scaleId, x+1, y)>centerValue &&
							get_value(octId, scaleId, x-1, y+1)>centerValue &&
							get_value(octId, scaleId, x, y+1)>centerValue &&
							get_value(octId, scaleId, x+1, y+1)>centerValue &&
							
							// next scale
							get_value(octId, scaleId+1, x-1, y-1)>centerValue &&
							get_value(octId, scaleId+1, x, y-1)>centerValue &&
							get_value(octId, scaleId+1, x+1, y-1)>centerValue &&
							get_value(octId, scaleId+1, x-1, y)>centerValue &&
							get_value(octId, scaleId+1, x, y)>centerValue &&
							get_value(octId, scaleId+1, x+1, y)>centerValue &&
							get_value(octId, scaleId+1, x-1, y+1)>centerValue &&
							get_value(octId, scaleId+1, x, y+1)>centerValue &&
							get_value(octId, scaleId+1, x+1, y+1)>centerValue)
						{
							float edgeScore= computeEdgeScore(octId, scaleId, x, y);
							if (edgeScore < (edgeThresh+1)*(edgeThresh+1)/edgeThresh)
							{
								scaleSpaceMax.push_back(ScaleSpaceMax(x, y, octId, scaleId));	
							}
						}
						}	
					}
				}
			}
		}
		return scaleSpaceMax;	
	}
};
	
float calcSigma(int octId, int scaleId)
{
	float sigmaNominal= 0.5;
	float sigma= 1.6f * powf(2, float(octId + float(scaleId) / 6));
	return sqrt(sigma*sigma - sigmaNominal*sigmaNominal);
}

void drawCircle(PVC_Image* img, int x, int y, float radius)
{	
	float* data= img->get_data();
	for (float t= 0; t<2*float(M_PI); t+= 0.001f)
	{
		int xx= x+ int(radius* sinf(t));
		int yy= y+ int(radius* cosf(t));
		if (xx>=0 && xx<img->w() && yy>=0 && yy<img->h())
		{
			data[yy*img->w()*4 + xx*4 + 0]= 0; 
			data[yy*img->w()*4 + xx*4 + 1]= 0; 
			data[yy*img->w()*4 + xx*4 + 2]= 1; 
		}
	}
}

void visualize(PVC_Image* img, vector<ScaleSpaceMax>& scaleSpaceMax)
{
	for (int i= 0; i<int(scaleSpaceMax.size()); i++)
	{
		float radius= calcSigma(scaleSpaceMax[i].octId, scaleSpaceMax[i].scaleId)*2;
		drawCircle(img, scaleSpaceMax[i].x, scaleSpaceMax[i].y, radius);
	}
}

int main( int argc, char *argv[] )
{


  PVC_Kernel   *gauss  = NULL;
  Filter_Image *img    = NULL;
  Filter_Image *result = NULL;

  double start, end;

  // load image
  char  filename[] = "../Data/Sunflowers_small.ppm"; 
  img = new OMP_Filter_Image();
  if ( img -> load( filename ) == 0 )
    fprintf( stderr, "The file '%s' has a size of %dx%d pixels ...\n", filename, img -> w(), img -> h() );
  else
    fprintf( stderr, "The file '%s' could not be loaded ...\n", filename );


  /////// compute DoG interest points

  start = omp_get_wtime();

  // construct scale-space
  ScaleSpace scaleSpace= ScaleSpace();
  scaleSpace.process(img);

  // calculate difference of gaussians
  DifferenceOfGaussians dog= DifferenceOfGaussians(0.015f, 0, 4);
  dog.process(scaleSpace);

  // find extrema
  vector<ScaleSpaceMax> scaleSpaceMax= dog.findMax();


  end = omp_get_wtime();

  // visualize
  visualize(img, scaleSpaceMax);
  img->save("ips.ppm");

  fprintf( stderr,"Scale Space: %0.2f s\n", (double)(end - start));

  // save debug images
  scaleSpace.saveImages("Output/scaleSpace_%d_%d.ppm");
  dog.saveImages("Output/dog_%d_%d.ppm");

  return 0;
}
